@@ -68,29 +68,43 @@ panfrost_alloc_variant(struct panfrost_uncompiled_shader *so)
    return util_dynarray_grow(&so->variants, struct panfrost_compiled_shader, 1);
 }
 
-static void
-lower_load_poly_line_smooth_enabled(nir_shader *nir,
-                                    const struct panfrost_shader_key *key)
+static bool
+lower_load_poly_line_smooth_enabled(nir_builder *b, nir_intrinsic_instr *intrin,
+                                    void *data)
 {
-   nir_function_impl *impl = nir_shader_get_entrypoint(nir);
-   nir_builder b = nir_builder_create(impl);
+   if (intrin->intrinsic != nir_intrinsic_load_poly_line_smooth_enabled)
+      return false;
 
-   nir_foreach_block_safe(block, impl) {
-      nir_foreach_instr_safe(instr, block) {
-         if (instr->type != nir_instr_type_intrinsic)
-            continue;
+   b->cursor = nir_before_instr(&intrin->instr);
+   nir_def_replace(&intrin->def, nir_imm_true(b));
+   return true;
+}
 
-         nir_intrinsic_instr *intrin = nir_instr_as_intrinsic(instr);
-         if (intrin->intrinsic != nir_intrinsic_load_poly_line_smooth_enabled)
-            continue;
+/* From the OpenGL 4.6 spec 14.3.1:
+ *
+ *    If MULTISAMPLE is disabled, multisample rasterization of all primitives
+ *    is equivalent to single-sample (fragment-center) rasterization, except
+ *    that the fragment coverage value is set to full coverage.
+ *
+ * So always use the original sample mask when multisample is disabled */
+static bool
+lower_sample_mask_writes(nir_builder *b, nir_intrinsic_instr *intrin,
+                         void *data)
+{
+   if (intrin->intrinsic != nir_intrinsic_store_output)
+      return false;
 
-         b.cursor = nir_before_instr(instr);
-         nir_def_rewrite_uses(&intrin->def, nir_imm_true(&b));
+   if (nir_intrinsic_io_semantics(intrin).location != FRAG_RESULT_SAMPLE_MASK)
+      return false;
 
-         nir_instr_remove(instr);
-         nir_instr_free(instr);
-      }
-   }
+   b->cursor = nir_before_instr(&intrin->instr);
+
+   nir_def *orig = nir_load_sample_mask(b);
+   nir_def *new = nir_b32csel(b, nir_load_multisampled_pan(b),
+                               intrin->src[0].ssa, orig);
+   nir_src_rewrite(&intrin->src[0], new);
+
+   return true;
 }
 
 static void
@@ -116,8 +130,13 @@ panfrost_shader_compile(struct panfrost_screen *screen, const nir_shader *ir,
    struct panfrost_compile_inputs inputs = {
       .debug = dbg,
       .gpu_id = panfrost_device_gpu_id(dev),
+      .push_uniforms = true,
    };
 
+   if (dev->arch >= 9)
+      /* Use LD_VAR_BUF for varying lookups. */
+      inputs.valhall.use_ld_var_buf = true;
+
    /* Lower this early so the backends don't have to worry about it */
    if (s->info.stage == MESA_SHADER_FRAGMENT) {
       inputs.fixed_varying_mask = key->fs.fixed_varying_mask;
@@ -128,10 +147,10 @@ panfrost_shader_compile(struct panfrost_screen *screen, const nir_shader *ir,
       inputs.no_idvs = s->info.has_transform_feedback_varyings;
 
       if (s->info.has_transform_feedback_varyings) {
-         NIR_PASS_V(s, nir_io_add_const_offset_to_base,
-                    nir_var_shader_in | nir_var_shader_out);
-         NIR_PASS_V(s, nir_io_add_intrinsic_xfb_info);
-         NIR_PASS_V(s, pan_lower_xfb);
+         NIR_PASS(_, s, nir_io_add_const_offset_to_base,
+                  nir_var_shader_in | nir_var_shader_out);
+         NIR_PASS(_, s, nir_io_add_intrinsic_xfb_info);
+         NIR_PASS(_, s, pan_lower_xfb);
       }
    }
 
@@ -139,37 +158,47 @@ panfrost_shader_compile(struct panfrost_screen *screen, const nir_shader *ir,
 
    if (s->info.stage == MESA_SHADER_FRAGMENT) {
       if (key->fs.nr_cbufs_for_fragcolor) {
-         NIR_PASS_V(s, panfrost_nir_remove_fragcolor_stores,
-                    key->fs.nr_cbufs_for_fragcolor);
+         NIR_PASS(_, s, panfrost_nir_remove_fragcolor_stores,
+                  key->fs.nr_cbufs_for_fragcolor);
       }
 
       if (key->fs.sprite_coord_enable) {
-         NIR_PASS_V(s, nir_lower_texcoord_replace_late,
-                    key->fs.sprite_coord_enable,
-                    true /* point coord is sysval */);
+         NIR_PASS(_, s, nir_lower_texcoord_replace_late,
+                  key->fs.sprite_coord_enable,
+                  true /* point coord is sysval */);
       }
 
       if (key->fs.clip_plane_enable) {
-         NIR_PASS_V(s, nir_lower_clip_fs, key->fs.clip_plane_enable, false);
+         NIR_PASS(_, s, nir_lower_clip_fs, key->fs.clip_plane_enable,
+                  false, true);
       }
 
       if (key->fs.line_smooth) {
-         NIR_PASS_V(s, nir_lower_poly_line_smooth, 16);
-         NIR_PASS_V(s, lower_load_poly_line_smooth_enabled, key);
-         NIR_PASS_V(s, nir_lower_alu);
+         NIR_PASS(_, s, nir_lower_poly_line_smooth, 16);
+         NIR_PASS(_, s, nir_shader_intrinsics_pass,
+                  lower_load_poly_line_smooth_enabled,
+                  nir_metadata_control_flow, key);
+         NIR_PASS(_, s, nir_lower_alu);
       }
+
+      NIR_PASS(_, s, nir_shader_intrinsics_pass,
+               lower_sample_mask_writes, nir_metadata_control_flow, NULL);
    }
 
    if (dev->arch <= 5 && s->info.stage == MESA_SHADER_FRAGMENT) {
-      NIR_PASS_V(s, pan_lower_framebuffer, key->fs.rt_formats,
-                 pan_raw_format_mask_midgard(key->fs.rt_formats), 0,
-                 panfrost_device_gpu_id(dev) < 0x700);
+      NIR_PASS(_, s, pan_lower_framebuffer, key->fs.rt_formats,
+               pan_raw_format_mask_midgard(key->fs.rt_formats), 0,
+               panfrost_device_gpu_id(dev) < 0x700);
    }
 
-   NIR_PASS_V(s, panfrost_nir_lower_sysvals, &out->sysvals);
+   if (s->info.stage == MESA_SHADER_VERTEX)
+      NIR_PASS(_, s, pan_nir_lower_static_noperspective,
+               key->vs.noperspective_varyings);
+
+   NIR_PASS(_, s, panfrost_nir_lower_sysvals, dev->arch, &out->sysvals);
 
    /* Lower resource indices */
-   NIR_PASS_V(s, panfrost_nir_lower_res_indices, &inputs);
+   NIR_PASS(_, s, panfrost_nir_lower_res_indices, &inputs);
 
    screen->vtbl.compile_shader(s, &inputs, &out->binary, &out->info);
 
@@ -233,16 +262,23 @@ panfrost_shader_get(struct pipe_screen *pscreen,
 }
 
 static void
-panfrost_build_key(struct panfrost_context *ctx,
-                   struct panfrost_shader_key *key,
-                   struct panfrost_uncompiled_shader *uncompiled)
+panfrost_build_vs_key(struct panfrost_context *ctx,
+                      struct panfrost_vs_key *key,
+                      struct panfrost_uncompiled_shader *uncompiled)
+{
+   struct panfrost_uncompiled_shader *fs = ctx->uncompiled[MESA_SHADER_FRAGMENT];
+
+   assert(fs != NULL && "too early");
+   key->noperspective_varyings = fs->noperspective_varyings;
+}
+
+static void
+panfrost_build_fs_key(struct panfrost_context *ctx,
+                      struct panfrost_fs_key *key,
+                      struct panfrost_uncompiled_shader *uncompiled)
 {
    const nir_shader *nir = uncompiled->nir;
 
-   /* We don't currently have vertex shader variants */
-   if (nir->info.stage != MESA_SHADER_FRAGMENT)
-      return;
-
    struct panfrost_device *dev = pan_device(ctx->base.screen);
    struct pipe_framebuffer_state *fb = &ctx->pipe_framebuffer;
    struct pipe_rasterizer_state *rast = (void *)ctx->rasterizer;
@@ -250,20 +286,20 @@ panfrost_build_key(struct panfrost_context *ctx,
 
    /* gl_FragColor lowering needs the number of colour buffers */
    if (uncompiled->fragcolor_lowered) {
-      key->fs.nr_cbufs_for_fragcolor = fb->nr_cbufs;
+      key->nr_cbufs_for_fragcolor = fb->nr_cbufs;
    }
 
    /* Point sprite lowering needed on Bifrost and newer */
    if (dev->arch >= 6 && rast && ctx->active_prim == MESA_PRIM_POINTS) {
-      key->fs.sprite_coord_enable = rast->sprite_coord_enable;
+      key->sprite_coord_enable = rast->sprite_coord_enable;
    }
 
    /* User clip plane lowering needed everywhere */
    if (rast) {
-      key->fs.clip_plane_enable = rast->clip_plane_enable;
+      key->clip_plane_enable = rast->clip_plane_enable;
 
       if (u_reduced_prim(ctx->active_prim) == MESA_PRIM_LINES)
-         key->fs.line_smooth = rast->line_smooth;
+         key->line_smooth = rast->line_smooth;
    }
 
    if (dev->arch <= 5) {
@@ -276,14 +312,33 @@ panfrost_build_key(struct panfrost_context *ctx,
          if (panfrost_blendable_formats_v6[fmt].internal)
             fmt = PIPE_FORMAT_NONE;
 
-         key->fs.rt_formats[i] = fmt;
+         key->rt_formats[i] = fmt;
       }
    }
 
    /* Funny desktop GL varying lowering on Valhall */
    if (dev->arch >= 9) {
       assert(vs != NULL && "too early");
-      key->fs.fixed_varying_mask = vs->fixed_varying_mask;
+      key->fixed_varying_mask = vs->fixed_varying_mask;
+   }
+}
+
+static void
+panfrost_build_key(struct panfrost_context *ctx,
+                   struct panfrost_shader_key *key,
+                   struct panfrost_uncompiled_shader *uncompiled)
+{
+   const nir_shader *nir = uncompiled->nir;
+
+   switch (nir->info.stage) {
+   case MESA_SHADER_VERTEX:
+      panfrost_build_vs_key(ctx, &key->vs, uncompiled);
+      break;
+   case MESA_SHADER_FRAGMENT:
+      panfrost_build_fs_key(ctx, &key->fs, uncompiled);
+      break;
+   default:
+      break;
    }
 }
 
@@ -331,7 +386,8 @@ panfrost_update_shader_variant(struct panfrost_context *ctx,
       return;
 
    /* We need linking information, defer this */
-   if (type == PIPE_SHADER_FRAGMENT && !ctx->uncompiled[PIPE_SHADER_VERTEX])
+   if ((type == PIPE_SHADER_FRAGMENT && !ctx->uncompiled[PIPE_SHADER_VERTEX]) ||
+       (type == PIPE_SHADER_VERTEX && !ctx->uncompiled[PIPE_SHADER_FRAGMENT]))
       return;
 
    /* Also defer, happens with GALLIUM_HUD */
@@ -377,6 +433,10 @@ static void
 panfrost_bind_fs_state(struct pipe_context *pctx, void *hwcso)
 {
    panfrost_bind_shader_state(pctx, hwcso, PIPE_SHADER_FRAGMENT);
+
+   /* Vertex shaders are linked with fragment shaders */
+   struct panfrost_context *ctx = pan_context(pctx);
+   panfrost_update_shader_variant(ctx, PIPE_SHADER_VERTEX);
 }
 
 static void *
@@ -408,7 +468,8 @@ panfrost_create_shader_state(struct pipe_context *pctx,
    if (nir->info.stage == MESA_SHADER_FRAGMENT &&
        nir->info.outputs_written & BITFIELD_BIT(FRAG_RESULT_COLOR)) {
 
-      NIR_PASS_V(nir, nir_lower_fragcolor, nir->info.fs.color_is_dual_source ? 1 : 8);
+      NIR_PASS(_, nir, nir_lower_fragcolor,
+               nir->info.fs.color_is_dual_source ? 1 : 8);
       so->fragcolor_lowered = true;
    }
 
@@ -416,13 +477,17 @@ panfrost_create_shader_state(struct pipe_context *pctx,
    struct panfrost_device *dev = pan_device(pctx->screen);
    pan_shader_preprocess(nir, panfrost_device_gpu_id(dev));
 
+   if (nir->info.stage == MESA_SHADER_FRAGMENT)
+      so->noperspective_varyings =
+         pan_nir_collect_noperspective_varyings_fs(nir);
+
    /* Vertex shaders get passed images through the vertex attribute descriptor
     * array. We need to add an offset to all image intrinsics so they point
     * to the right attribute.
     */
    if (nir->info.stage == MESA_SHADER_VERTEX && dev->arch <= 7) {
-      NIR_PASS_V(nir, pan_lower_image_index,
-                 util_bitcount64(nir->info.inputs_read));
+      NIR_PASS(_, nir, pan_lower_image_index,
+               util_bitcount64(nir->info.inputs_read));
    }
 
    /* If this shader uses transform feedback, compile the transform
@@ -432,7 +497,7 @@ panfrost_create_shader_state(struct pipe_context *pctx,
 
    if (so->nir->xfb_info) {
       so->xfb = calloc(1, sizeof(struct panfrost_compiled_shader));
-      so->xfb->key.vs_is_xfb = true;
+      so->xfb->key.vs.is_xfb = true;
 
       panfrost_shader_get(ctx->base.screen, &ctx->shaders, &ctx->descs, so,
                           &ctx->base.debug, so->xfb, 0);
