@@ -51,7 +51,6 @@
 #include "compiler/glsl/ir.h"
 #include "compiler/glsl/ir_optimization.h"
 #include "compiler/glsl/linker_util.h"
-#include "compiler/glsl/program.h"
 #include "compiler/glsl/shader_cache.h"
 #include "compiler/glsl/string_to_uint_map.h"
 
@@ -63,71 +62,6 @@ type_size(const struct glsl_type *type)
    return glsl_count_attribute_slots(type, false);
 }
 
-/* Depending on PIPE_CAP_TGSI_TEXCOORD (st->needs_texcoord_semantic) we
- * may need to fix up varying slots so the glsl->nir path is aligned
- * with the anything->tgsi->nir path.
- */
-static void
-st_nir_fixup_varying_slots(struct st_context *st, nir_shader *shader,
-                           nir_variable_mode mode)
-{
-   if (st->needs_texcoord_semantic)
-      return;
-
-   /* This is called from finalize, but we don't want to do this adjustment twice. */
-   assert(!st->allow_st_finalize_nir_twice);
-
-   nir_foreach_variable_with_modes(var, shader, mode) {
-      if (var->data.location >= VARYING_SLOT_VAR0 && var->data.location < VARYING_SLOT_PATCH0) {
-         var->data.location += 9;
-      } else if (var->data.location == VARYING_SLOT_PNTC) {
-         var->data.location = VARYING_SLOT_VAR8;
-      } else if ((var->data.location >= VARYING_SLOT_TEX0) &&
-               (var->data.location <= VARYING_SLOT_TEX7)) {
-         var->data.location += VARYING_SLOT_VAR0 - VARYING_SLOT_TEX0;
-      }
-   }
-}
-
-/* input location assignment for VS inputs must be handled specially, so
- * that it is aligned w/ st's vbo state.
- * (This isn't the case with, for ex, FS inputs, which only need to agree
- * on varying-slot w/ the VS outputs)
- */
-void
-st_nir_assign_vs_in_locations(struct nir_shader *nir)
-{
-   if (nir->info.stage != MESA_SHADER_VERTEX || nir->info.io_lowered)
-      return;
-
-   nir->num_inputs = util_bitcount64(nir->info.inputs_read);
-
-   bool removed_inputs = false;
-
-   nir_foreach_shader_in_variable_safe(var, nir) {
-      /* NIR already assigns dual-slot inputs to two locations so all we have
-       * to do is compact everything down.
-       */
-      if (nir->info.inputs_read & BITFIELD64_BIT(var->data.location)) {
-         var->data.driver_location =
-            util_bitcount64(nir->info.inputs_read &
-                              BITFIELD64_MASK(var->data.location));
-      } else {
-         /* Convert unused input variables to shader_temp (with no
-          * initialization), to avoid confusing drivers looking through the
-          * inputs array and expecting to find inputs with a driver_location
-          * set.
-          */
-         var->data.mode = nir_var_shader_temp;
-         removed_inputs = true;
-      }
-   }
-
-   /* Re-lower global vars, to deal with any dead VS inputs. */
-   if (removed_inputs)
-      NIR_PASS(_, nir, nir_lower_global_vars_to_local);
-}
-
 static int
 st_nir_lookup_parameter_index(struct gl_program *prog, nir_variable *var)
 {
@@ -324,11 +258,10 @@ st_glsl_to_nir_post_opts(struct st_context *st, struct gl_program *prog,
        !st->ctx->Const.PackedDriverUniformStorage)
       NIR_PASS(_, nir, st_nir_lower_builtin);
 
-   if (!screen->get_param(screen, PIPE_CAP_NIR_ATOMICS_AS_DEREF))
+   if (!screen->caps.nir_atomics_as_deref)
       NIR_PASS(_, nir, gl_nir_lower_atomics, shader_program, true);
 
    NIR_PASS(_, nir, nir_opt_intrinsics);
-   NIR_PASS(_, nir, nir_opt_fragdepth);
 
    /* Lower 64-bit ops. */
    if (nir->options->lower_int64_options ||
@@ -369,7 +302,7 @@ st_glsl_to_nir_post_opts(struct st_context *st, struct gl_program *prog,
       nir_var_shader_in | nir_var_shader_out | nir_var_function_temp;
    nir_remove_dead_variables(nir, mask, NULL);
 
-   if (!st->has_hw_atomics && !screen->get_param(screen, PIPE_CAP_NIR_ATOMICS_AS_DEREF)) {
+   if (!st->has_hw_atomics && !screen->caps.nir_atomics_as_deref) {
       unsigned align_offset_state = 0;
       if (st->ctx->Const.ShaderStorageBufferOffsetAlignment > 4) {
          struct gl_program_parameter_list *params = prog->Parameters;
@@ -383,13 +316,15 @@ st_glsl_to_nir_post_opts(struct st_context *st, struct gl_program *prog,
    }
 
    st_set_prog_affected_state_flags(prog);
-
-   st_finalize_nir_before_variants(nir);
+   nir_shader_gather_info(nir, nir_shader_get_entrypoint(nir));
 
    char *msg = NULL;
    if (st->allow_st_finalize_nir_twice) {
       st_serialize_base_nir(prog, nir);
-      msg = st_finalize_nir(st, prog, shader_program, nir, true, true, false);
+      st_finalize_nir(st, prog, shader_program, nir, true, false);
+
+      if (screen->finalize_nir)
+         msg = screen->finalize_nir(screen, nir);
    }
 
    if (st->ctx->_Shader->Flags & GLSL_DUMP) {
@@ -404,45 +339,6 @@ st_glsl_to_nir_post_opts(struct st_context *st, struct gl_program *prog,
    return msg;
 }
 
-static void
-st_nir_vectorize_io(nir_shader *producer, nir_shader *consumer)
-{
-   if (consumer)
-      NIR_PASS(_, consumer, nir_lower_io_to_vector, nir_var_shader_in);
-
-   if (!producer)
-      return;
-
-   NIR_PASS(_, producer, nir_lower_io_to_vector, nir_var_shader_out);
-
-   if (producer->info.stage == MESA_SHADER_TESS_CTRL &&
-       producer->options->vectorize_tess_levels)
-      NIR_PASS(_, producer, nir_vectorize_tess_levels);
-
-   NIR_PASS(_, producer, nir_opt_combine_stores, nir_var_shader_out);
-
-   if ((producer)->info.stage != MESA_SHADER_TESS_CTRL) {
-      /* Calling lower_io_to_vector creates output variable writes with
-       * write-masks.  We only support these for TCS outputs, so for other
-       * stages, we need to call nir_lower_io_to_temporaries to get rid of
-       * them.  This, in turn, creates temporary variables and extra
-       * copy_deref intrinsics that we need to clean up.
-       */
-      NIR_PASS(_, producer, nir_lower_io_to_temporaries,
-                 nir_shader_get_entrypoint(producer), true, false);
-      NIR_PASS(_, producer, nir_lower_global_vars_to_local);
-      NIR_PASS(_, producer, nir_split_var_copies);
-      NIR_PASS(_, producer, nir_lower_var_copies);
-   }
-
-   /* Undef scalar store_deref intrinsics are not ignored by nir_lower_io,
-    * so they must be removed before that. These passes remove them.
-    */
-   NIR_PASS(_, producer, nir_lower_vars_to_ssa);
-   NIR_PASS(_, producer, nir_opt_undef);
-   NIR_PASS(_, producer, nir_opt_dce);
-}
-
 extern "C" {
 
 bool
@@ -465,17 +361,13 @@ st_nir_lower_wpos_ytransform(struct nir_shader *nir,
    memcpy(wpos_options.state_tokens, wposTransformState,
           sizeof(wpos_options.state_tokens));
    wpos_options.fs_coord_origin_upper_left =
-      pscreen->get_param(pscreen,
-                         PIPE_CAP_FS_COORD_ORIGIN_UPPER_LEFT);
+      pscreen->caps.fs_coord_origin_upper_left;
    wpos_options.fs_coord_origin_lower_left =
-      pscreen->get_param(pscreen,
-                         PIPE_CAP_FS_COORD_ORIGIN_LOWER_LEFT);
+      pscreen->caps.fs_coord_origin_lower_left;
    wpos_options.fs_coord_pixel_center_integer =
-      pscreen->get_param(pscreen,
-                         PIPE_CAP_FS_COORD_PIXEL_CENTER_INTEGER);
+      pscreen->caps.fs_coord_pixel_center_integer;
    wpos_options.fs_coord_pixel_center_half_integer =
-      pscreen->get_param(pscreen,
-                         PIPE_CAP_FS_COORD_PIXEL_CENTER_HALF_INTEGER);
+      pscreen->caps.fs_coord_pixel_center_half_integer;
 
    if (nir_lower_wpos_ytransform(nir, &wpos_options)) {
       _mesa_add_state_reference(prog->Parameters, wposTransformState);
@@ -545,9 +437,6 @@ st_link_glsl_to_nir(struct gl_context *ctx,
             prog->nir->info.label = ralloc_strdup(shader, shader_program->Label);
       }
 
-      memcpy(prog->nir->info.source_blake3, shader->linked_source_blake3,
-             BLAKE3_OUT_LEN);
-
       nir_shader_gather_info(prog->nir, nir_shader_get_entrypoint(prog->nir));
       if (!st->ctx->SoftFP64 && ((prog->nir->info.bit_sizes_int | prog->nir->info.bit_sizes_float) & 64) &&
           (options->lower_doubles_options & nir_lower_fp64_full_software) != 0) {
@@ -586,27 +475,18 @@ st_link_glsl_to_nir(struct gl_context *ctx,
       const struct gl_shader_compiler_options *options =
             &ctx->Const.ShaderCompilerOptions[stage];
 
-      if (nir->info.io_lowered) {
-         /* Since IO is lowered, we won't need the IO variables from now on.
-          * nir_build_program_resource_list was the last pass that needed them.
-          */
-         NIR_PASS_V(nir, nir_remove_dead_variables,
-                    nir_var_shader_in | nir_var_shader_out, NULL);
-      }
+      /* Since IO is lowered, we won't need the IO variables from now on.
+       * nir_build_program_resource_list was the last pass that needed them.
+       */
+      NIR_PASS_V(nir, nir_remove_dead_variables,
+                 nir_var_shader_in | nir_var_shader_out, NULL);
 
       /* If there are forms of indirect addressing that the driver
        * cannot handle, perform the lowering pass.
        */
-      if (options->EmitNoIndirectInput || options->EmitNoIndirectOutput ||
-          options->EmitNoIndirectTemp || options->EmitNoIndirectUniform) {
+      if (options->EmitNoIndirectTemp || options->EmitNoIndirectUniform) {
          nir_variable_mode mode = (nir_variable_mode)0;
 
-         if (!nir->info.io_lowered) {
-            mode |= options->EmitNoIndirectInput ?
-               nir_var_shader_in : (nir_variable_mode)0;
-            mode |= options->EmitNoIndirectOutput ?
-               nir_var_shader_out : (nir_variable_mode)0;
-         }
          mode |= options->EmitNoIndirectTemp ?
             nir_var_function_temp : (nir_variable_mode)0;
          mode |= options->EmitNoIndirectUniform ?
@@ -629,50 +509,6 @@ st_link_glsl_to_nir(struct gl_context *ctx,
       struct nir_lower_compute_system_values_options cs_options = {};
       NIR_PASS(_, nir, nir_lower_system_values);
       NIR_PASS(_, nir, nir_lower_compute_system_values, &cs_options);
-
-      if (nir->info.io_lowered)
-         continue; /* the rest is for non-lowered IO only */
-
-      /* Remap the locations to slots so those requiring two slots will occupy
-       * two locations. For instance, if we have in the IR code a dvec3 attr0 in
-       * location 0 and vec4 attr1 in location 1, in NIR attr0 will use
-       * locations/slots 0 and 1, and attr1 will use location/slot 2
-       */
-      if (nir->info.stage == MESA_SHADER_VERTEX && !shader_program->data->spirv)
-         nir_remap_dual_slot_attributes(nir, &shader->Program->DualSlotInputs);
-
-      if (i >= 1) {
-         struct gl_program *prev_shader = linked_shader[i - 1]->Program;
-
-         /* We can't use nir_compact_varyings with transform feedback, since
-          * the pipe_stream_output->output_register field is based on the
-          * pre-compacted driver_locations.
-          */
-         if (!(prev_shader->sh.LinkedTransformFeedback &&
-               prev_shader->sh.LinkedTransformFeedback->NumVarying > 0))
-            nir_compact_varyings(prev_shader->nir,
-                                 nir, ctx->API != API_OPENGL_COMPAT);
-
-         if (ctx->Const.ShaderCompilerOptions[shader->Stage].NirOptions->vectorize_io)
-            st_nir_vectorize_io(prev_shader->nir, nir);
-      }
-   }
-
-   /* If the program is a separate shader program check if we need to vectorise
-    * the first and last program interfaces too.
-    */
-   if (shader_program->SeparateShader && num_shaders > 0) {
-      struct gl_linked_shader *first_shader = linked_shader[0];
-      struct gl_linked_shader *last_shader = linked_shader[num_shaders - 1];
-      if (first_shader->Stage != MESA_SHADER_COMPUTE) {
-         if (ctx->Const.ShaderCompilerOptions[first_shader->Stage].NirOptions->vectorize_io &&
-             first_shader->Stage > MESA_SHADER_VERTEX)
-            st_nir_vectorize_io(NULL, first_shader->Program->nir);
-
-         if (ctx->Const.ShaderCompilerOptions[last_shader->Stage].NirOptions->vectorize_io &&
-             last_shader->Stage < MESA_SHADER_FRAGMENT)
-            st_nir_vectorize_io(last_shader->Program->nir, NULL);
-      }
    }
 
    struct shader_info *prev_info = NULL;
@@ -737,18 +573,8 @@ st_link_glsl_to_nir(struct gl_context *ctx,
       prog->info.num_abos = old_info.num_abos;
 
       if (prog->info.stage == MESA_SHADER_VERTEX) {
-         if (prog->nir->info.io_lowered) {
-            prog->info.inputs_read = prog->nir->info.inputs_read;
-            prog->DualSlotInputs = prog->nir->info.dual_slot_inputs;
-         } else {
-            /* NIR expands dual-slot inputs out to two locations.  We need to
-             * compact things back down GL-style single-slot inputs to avoid
-             * confusing the state tracker.
-             */
-            prog->info.inputs_read =
-               nir_get_single_slot_attribs_mask(prog->nir->info.inputs_read,
-                                                prog->DualSlotInputs);
-         }
+         prog->info.inputs_read = prog->nir->info.inputs_read;
+         prog->DualSlotInputs = prog->nir->info.dual_slot_inputs;
 
          /* Initialize st_vertex_program members. */
          st_prepare_vertex_program(prog);
@@ -763,7 +589,13 @@ st_link_glsl_to_nir(struct gl_context *ctx,
       st_store_nir_in_disk_cache(st, prog);
 
       st_release_variants(st, prog);
-      st_finalize_program(st, prog);
+      char *error = st_finalize_program(st, prog, true);
+
+      if (error) {
+         linker_error(shader_program, error);
+         free(error);
+         return false;
+      }
    }
 
    struct pipe_context *pctx = st_context(ctx)->pipe;
@@ -788,51 +620,12 @@ st_link_glsl_to_nir(struct gl_context *ctx,
    return true;
 }
 
-void
-st_nir_assign_varying_locations(struct st_context *st, nir_shader *nir)
-{
-   /* Lowered IO don't have variables, so exit. */
-   if (nir->info.io_lowered)
-      return;
-
-   if (nir->info.stage == MESA_SHADER_VERTEX) {
-      nir_assign_io_var_locations(nir, nir_var_shader_out,
-                                  &nir->num_outputs,
-                                  nir->info.stage);
-      st_nir_fixup_varying_slots(st, nir, nir_var_shader_out);
-   } else if (nir->info.stage == MESA_SHADER_GEOMETRY ||
-              nir->info.stage == MESA_SHADER_TESS_CTRL ||
-              nir->info.stage == MESA_SHADER_TESS_EVAL) {
-      nir_assign_io_var_locations(nir, nir_var_shader_in,
-                                  &nir->num_inputs,
-                                  nir->info.stage);
-      st_nir_fixup_varying_slots(st, nir, nir_var_shader_in);
-
-      nir_assign_io_var_locations(nir, nir_var_shader_out,
-                                  &nir->num_outputs,
-                                  nir->info.stage);
-      st_nir_fixup_varying_slots(st, nir, nir_var_shader_out);
-   } else if (nir->info.stage == MESA_SHADER_FRAGMENT) {
-      nir_assign_io_var_locations(nir, nir_var_shader_in,
-                                  &nir->num_inputs,
-                                  nir->info.stage);
-      st_nir_fixup_varying_slots(st, nir, nir_var_shader_in);
-      nir_assign_io_var_locations(nir, nir_var_shader_out,
-                                  &nir->num_outputs,
-                                  nir->info.stage);
-   } else if (nir->info.stage == MESA_SHADER_COMPUTE) {
-       /* TODO? */
-   } else {
-      unreachable("invalid shader type");
-   }
-}
-
 void
 st_nir_lower_samplers(struct pipe_screen *screen, nir_shader *nir,
                       struct gl_shader_program *shader_program,
                       struct gl_program *prog)
 {
-   if (screen->get_param(screen, PIPE_CAP_NIR_SAMPLERS_AS_DEREF))
+   if (screen->caps.nir_samplers_as_deref)
       NIR_PASS(_, nir, gl_nir_lower_samplers_as_deref, shader_program);
    else
       NIR_PASS(_, nir, gl_nir_lower_samplers, shader_program);
@@ -881,22 +674,17 @@ st_nir_lower_uniforms(struct st_context *st, nir_shader *nir)
 /* Last third of preparing nir from glsl, which happens after shader
  * variant lowering.
  */
-char *
+void
 st_finalize_nir(struct st_context *st, struct gl_program *prog,
-                struct gl_shader_program *shader_program,
-                nir_shader *nir, bool finalize_by_driver,
-                bool is_before_variants,
-                bool is_draw_shader)
+                struct gl_shader_program *shader_program, nir_shader *nir,
+                bool is_before_variants, bool is_draw_shader)
 {
    struct pipe_screen *screen = st->screen;
 
    MESA_TRACE_FUNC();
 
-   NIR_PASS(_, nir, nir_split_var_copies);
-   NIR_PASS(_, nir, nir_lower_var_copies);
-
    const bool lower_tg4_offsets =
-      !is_draw_shader && !st->screen->get_param(screen, PIPE_CAP_TEXTURE_GATHER_OFFSETS);
+      !is_draw_shader && !st->screen->caps.texture_gather_offsets;
 
    if (!is_draw_shader && (st->lower_rect_tex || lower_tg4_offsets)) {
       struct nir_lower_tex_options opts = {0};
@@ -906,7 +694,6 @@ st_finalize_nir(struct st_context *st, struct gl_program *prog,
       NIR_PASS(_, nir, nir_lower_tex, &opts);
    }
 
-   st_nir_assign_varying_locations(st, nir);
    st_nir_assign_uniform_locations(st->ctx, prog, nir);
 
    /* Set num_uniforms in number of attribute slots (vec4s) */
@@ -923,14 +710,8 @@ st_finalize_nir(struct st_context *st, struct gl_program *prog,
    }
 
    st_nir_lower_samplers(screen, nir, shader_program, prog);
-   if (!is_draw_shader && !screen->get_param(screen, PIPE_CAP_NIR_IMAGES_AS_DEREF))
+   if (!is_draw_shader && !screen->caps.nir_images_as_deref)
       NIR_PASS(_, nir, gl_nir_lower_images, false);
-
-   char *msg = NULL;
-   if (!is_draw_shader && finalize_by_driver && screen->finalize_nir)
-      msg = screen->finalize_nir(screen, nir);
-
-   return msg;
 }
 
 /**
